// ******************************************************
//  file    HLsupport.h
//  purpose The class HLsupport is used by sdm.cpp to 
//			facilitate lookup/insertion/deletion of 
//			active locations
//  author  Alexandre Bouchard
//  written May 2003
// ******************************************************

#ifndef HLSUPPORT_H
#define HLSUPPORT_H

/*
	A hashtable (sgi implementation - http://www.sgi.com/tech/stl/index.html) 
	is needed for the class HLsupport.
*/
#include <ext/hash_map> // not std (yet) with STL, but included by default with gcc 3.x
#include <cstring>
#include <ext/slist>

namespace Sgi = ::__gnu_cxx;

/*
	The default comparator of char* is the address of the pointers!
	We will need the natural definition of equality of string in our hash
*/
struct eqstr
{
	bool operator()(const char* s1, const char* s2) const
	{
		return strcmp(s1, s2) == 0;
	}
};

// We cannot include sdm.h because it will be a circular import
class List;
class State;

typedef Sgi::slist<State*> ListOfStates;
typedef Sgi::hash_map<const char*, List*, Sgi::hash<const char*>, eqstr> HashMap;

/*
	See implementation (HLsupport.cpp) for more information and comments
 */
class HLsupport
{
public:
			HLsupport(int dim, double* radius, double* leftBounds, double* rightBounds);
			~HLsupport();
	void	addHL(const State& s, int ind);
	void	removeHL(const State& s, int ind);
	List*   potentiallyActiveSet(const State& s);
	double* LeftBounds;
	double* RightBounds;
	
	//private:
	HashMap hashMap;
	int		numberOfDimensions;
	double* Radius;  
	char*   computeKey(const State& state);
	ListOfStates* getNeighborCubesRepresentatives(const State& state, int currentDimension = -1);
};

// Test cases declaration
//int main();
void testHashTable();
void testComputeKey();
void testGetNeighborCubesRepresentatives();
void testAddHL();
void printStringAsIntegers(char* string);
void printCoordinates(const State& state);

#endif
