#include <stdlib.h>
#include <string.h>
#include<stdio.h>
#include<math.h>
#include<sys/types.h>		
#include<iostream.h>
#include<fstream.h>


#include "interface_classes.h"
#include "estimators.h"

Signal::Signal(){}
Signal::~Signal(){}


ContinuousSignal::ContinuousSignal(int in, double* left, double* right, int* nn){
  int i;

  Inputs=in;
  LeftBounds=new double[Inputs];
  RightBounds=new double[Inputs];
  n=new int[Inputs];
  IndCoef=new int[Inputs];
  value=new double[Inputs];

  N=1;
  IndCoef[Inputs-1]=1;
  for(i=Inputs-1; i>=0; i--){
    LeftBounds[i]=left[i];
    RightBounds[i]=right[i];
    n[i]=nn[i];
    value[i]=0;
    N=N*n[i];
    if (i<(Inputs-1))
      IndCoef[i]=IndCoef[i+1]*n[i+1];
  }

}

int ContinuousSignal::Index(){

  int i, ind;
  int index=0;
  for(i=0; i<Inputs; i++){
    if (value[i]>=RightBounds[i]){ 
      ind=n[i]-1; 
    }
    else{

      ind=(int)(float)((value[i]-LeftBounds[i])/((RightBounds[i]-LeftBounds[i])/(double)(n[i]))); 
      if (ind == n[i]) ind=n[i]-1;
    }
    index=index+IndCoef[i]*ind;
  }
		
  if ((index<0) || (index>=N)) {
    cout << "Error (ContinuousSignal): bin index out of limits" << endl;
    cout << "index=" << index << endl;
    cout << "N=" << N << endl;
    for(i=0; i<Inputs; i++)
      cout << "value[" << i << "]=" << value[i] << " n[" << i << "]=" << n[i] << endl;
    exit(EXIT_FAILURE);
  }

  return index;

}

void ContinuousSignal::setValue(double* v){
  int i;

  for (i=0; i<Inputs; i++)
    value[i]=v[i];
}

ContinuousSignal::~ContinuousSignal(){

  delete [] value;
  delete [] LeftBounds;
  delete [] RightBounds;
  delete [] n;
  delete [] IndCoef;
 
}

/////////////////////////////////////////////////


///////////////////////////////////////////////////////

EntropyEstimator::EntropyEstimator(int type, int n, double volume, double l){
  int i;

  Type=type;
  N=n;
  Volume=volume;

  if ((Type==0) && (Volume==0)){
    cout << "Error (EntropyEstmator) : please specify the volume of the discretization bin for the continuous signal." << endl;
    exit(EXIT_FAILURE);
  }

  Samples=0;
  entropy=0;
  //for (i=0; i<N; i++)
  //entropy-=1/(double)N*log2(1/(double)N);

  entropy=0;
  entropy=entropy+log2(Volume);

  counts=new int[N];

  if (counts==NULL){
    cout << "EntropyEstimator : Memory allocation error (counts)" << endl;
    exit(EXIT_FAILURE);
  }
  for(i=0; i<N; i++)
    counts[i]=0;

  lambda=l;
  powerLambda=1;

}

void EntropyEstimator::update(Signal* sig){
  int index;

  index=sig->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimator (update) -  index out of bounds" << endl;
    exit(EXIT_FAILURE);
  }
  
  if (Samples>0)
  
    {
      if (lambda<1){
	powerLambda=powerLambda*lambda;
	entropy=(powerLambda-lambda)/(powerLambda-1)*entropy+(lambda-1)/(powerLambda-1)*(g(Samples)-g(counts[index])+log2(Volume));
      }
      else
	entropy=Samples/(Samples+1)*entropy+1/(Samples+1)*(g(Samples)-g(counts[index])+log2(Volume));
    
    }
  else 
    if (lambda>0.999) entropy=0;
  counts[index]++;
  Samples++;

}

double EntropyEstimator::g(double x){

  return (x+1)*log2(x+1)-x*log2(x);

}

double EntropyEstimator::log2(double x){

  if (x==0) return 0;
  return (double)(log10(x))/(double)(log10(2.0));

}

int EntropyEstimator::getCount(Signal* sig){

  int index;

  index=sig->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection -  index out of bounds " << index << endl;
    exit(EXIT_FAILURE);
  }
  return counts[index];


}

double EntropyEstimator::getProbability(Signal* sig){

  int index;

  if (Samples==0) return 0;

  index=sig->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimator (getProbability) -  index out of bounds "<< index << endl;
    exit(EXIT_FAILURE);
  }
  if (Type==1)
    return counts[index]/(Samples);
  else
    return counts[index]/(Volume*Samples);

}

bool EntropyEstimator::getProbabilityVector(double* p){

  int i;
  if (Samples==0){
    for (i=0; i<N; i++)
      p[i]=0;
    return false;
  }
  else{
    for (i=0; i<N; i++)
      p[i]=(double)counts[i]/Samples;
    return true;
  }
}

double EntropyEstimator::getEntropy(){

  return entropy;

}

int EntropyEstimator::getNumberOfSamples(){

  return (int)Samples;

}

EntropyEstimator::~EntropyEstimator(){
  
  delete [] counts;

}


//////////////////////////////////////////////////////////////////////

EntropyEstimatorCollection::EntropyEstimatorCollection(int EstimatorsNumber, int type, int EstimatorBins, double volume, double l){

  int i;
  N=EstimatorsNumber;
  
  items = new EntropyEstimator*[N];
  if (items == NULL){
    cout << "EntropyEstimatorCollection: Memory allocation problem" << endl;
    exit(EXIT_FAILURE);
  }

  for(i=0; i<N; i++){
    items[i]= new EntropyEstimator(type, EstimatorBins, volume, l);
    if (items[i]==NULL){
      cout << "EntropyEstimatorCollection : Memory allocation error" << endl;
      exit(EXIT_FAILURE);
    }
  }

}

void EntropyEstimatorCollection::update(Signal* item, Signal* sample){

  int index;
  index=item->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection (update)-  index out of bounds " << index << endl;
    exit(EXIT_FAILURE);
  }
  items[index]->update(sample);

}

int EntropyEstimatorCollection::getCount(Signal* item, Signal* fromBin ){
  
  int index;
  index=item->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection (getCount)-  index out of bounds " << index << endl;
    exit(EXIT_FAILURE);
  }
  return items[index]->getCount(fromBin);

}

bool EntropyEstimatorCollection::getProbabilityVector(Signal* item, double* p){

  int index;
  index=item->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection (getProbabilityVector)-  index out of bounds " << index << endl;
    exit(EXIT_FAILURE);
  }
  return items[index]->getProbabilityVector(p);
}

double EntropyEstimatorCollection::getProbability(Signal* item, Signal* fromBin){

  int index;
  index=item->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection (getProbability)-  index out of bounds " << index << endl;
    exit(EXIT_FAILURE);
  }
  return items[index]->getProbability(fromBin);

}

double EntropyEstimatorCollection::getEntropy(Signal* item){
  
  int index;
  index=item->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection (getEntropy)-  index out of bounds " << index << endl;
    exit(EXIT_FAILURE);
  }
    return items[index]->getEntropy();

}

int EntropyEstimatorCollection::getNumberOfSamples(Signal* item){

  int index;
  index=item->Index();
  if ((index<0) || (index >= N)){
    cout << "Error: EntropyEstimatorCollection -  index out of bounds" << endl;
    exit(EXIT_FAILURE);
  }
  return items[index]->getNumberOfSamples();
}

void EntropyEstimatorCollection::print(){
 
  int i;
  for (i=0; i<N; i++)
    cout << items[i]->getEntropy() << endl;

}

EntropyEstimatorCollection::~EntropyEstimatorCollection(){

  int i;
  for(i=0; i<N; i++)
    delete items[i];

  delete [] items;

}
/////////////////////////////////////////////////////////////

stdEstimatorCollection::stdEstimatorCollection(int EstimatorsNumber){

  int i;
  N=EstimatorsNumber;

  means=new double[N];
  squares=new double[N];
  Samples=new int[N];

  for (i=0; i<N; i++){
    means[i]=0;
    squares[i]=0;
    Samples[i]=0;

  }

}


void stdEstimatorCollection::update(Signal* item, double sample){
  int index;
  index=item->Index();

  //update mean
  means[index]=(Samples[index]*means[index]+sample)/(double)(Samples[index]+1);

  //update squares
  squares[index]+=sample*sample;
  
  //update stdnumber of samples

  Samples[index]++;

}

double stdEstimatorCollection::getSTD(Signal* item){
  int index;
  index=item->Index();
  
  if (Samples[index]<2)
    return 0;
  else
    return (squares[index]-Samples[index]*means[index]*means[index])/(double)(Samples[index]-1);
}


int stdEstimatorCollection::getNumberOfSamples(Signal* item){
  int index;
  index=item->Index();
  return Samples[index];
}

stdEstimatorCollection::~stdEstimatorCollection(){
  delete [] squares;
  delete [] means;
  delete [] Samples;
}
////////////////////////////////////////////////////////////

MeanEstimatorCollection::MeanEstimatorCollection(int EstimatorsNumber){

  int i;
  N=EstimatorsNumber;

  means=new double[N];
  Samples=new int[N];

  for(i=0;i< N;i++){
    means[i]=0.0;
    Samples[i]=0;
  }

}

void MeanEstimatorCollection::update(Signal* item, double sample){

  int index;
  index=item->Index();
  means[index]=(Samples[index]*means[index]+sample)/(double)(Samples[index]+1);

  Samples[index]++;

}

int MeanEstimatorCollection::getNumberOfSamples(Signal* item){

  int index;
  index=item->Index();

  return Samples[index];

}

double MeanEstimatorCollection::getMean(Signal* item){
  
  int index;
  index=item->Index();

  return means[index];
}

MeanEstimatorCollection::~MeanEstimatorCollection(){

  delete [] means;
  delete [] Samples;

}

///////////////////////////////////////////////////////

ForgettingMeanSTDEstimatorCollection::ForgettingMeanSTDEstimatorCollection(int EstimatorsNumber, double l, double STDmax) : MeanEstimatorCollection(EstimatorsNumber){
  lambda=l;
  lambdaPower=1;
  STDs=new double[N];
  int i;
  for (i=0; i<N; i++)
    STDs[i]=STDmax*STDmax;
}


void ForgettingMeanSTDEstimatorCollection::update(Signal* item, double sample){
  
  int index;
  index=item->Index();
  lambdaPower=lambdaPower*lambda;
  means[index]=(lambdaPower-lambda)/(lambdaPower-1)*means[index]+(lambda-1)/(lambdaPower-1)*sample;

  STDs[index]=(lambdaPower/lambda-lambda)/(lambdaPower/lambda-1)*STDs[index]+(lambda-1)/(lambdaPower/lambda-1)*(sample-means[index])*(sample-means[index]);

  Samples[index]++;
}

void ForgettingMeanSTDEstimatorCollection::get(Signal* item, double& m, double& d){
  
  int index;
  index=item->Index();

  m=means[index];

  if (STDs[index]<0)
    d=0;
  else
    d=sqrt(STDs[index]);

}

ForgettingMeanSTDEstimatorCollection::~ForgettingMeanSTDEstimatorCollection()
{
  delete [] STDs;
}

///////////////////////////////////////////////////////

StorageTable::StorageTable(int items){

  Items=items;
  values = new double[Items];

  if (values == NULL){
    cout << "Error (StorageTable): memory allocation failed " << endl;
    exit(EXIT_FAILURE);
  }

  int i;
  for (i=0; i<Items; i++)
    values[i]=0.0;
}

void StorageTable::storeItem(Signal* item, double value){

  int index;

  index=item->Index();
  values[index]=value;

}

double StorageTable::getItem(Signal* item){

  int index;

  index=item->Index();
  return values[index];

}

double StorageTable::getMean(){
  int i;
  double mean=0;

  for (i=0; i<Items; i++)
    mean+=values[i];

  if (Items!=0)
    return mean/(double)Items;
  else
    return 0;
}

double StorageTable::getSTD(){
  int i;
  double std;
  double mean;

  mean=getMean();
  std=0;
  for(i=0; i<Items; i++)
    std+=(values[i]-mean)*(values[i]-mean);

  std/=(double)(Items-1);
  std=sqrt(std);

  return std;
}

StorageTable::~StorageTable(){

  delete [] values;

}

void StorageTable::saveTable(char* fileName){

  ofstream file(fileName);
  if (file.fail()){
    cout << "Error: cannot open file " << fileName << endl;
    exit(EXIT_FAILURE);
  }

  int i;
  for(i=0; i<Items; i++)
    file << values[i] << endl;
  
}


void StorageTable::loadTable(char* fileName){

  ifstream file(fileName);
  if (file.fail()){
    cout << "Error: cannot open file " << fileName << endl;
    exit(EXIT_FAILURE);
  }

  int i;
  char c;

  for(i=0; i<Items; i++){
    file >> values[i];
    file.get(c); //get eol
    if (file.fail()){
      cout << "Error reading file " << fileName << endl;
      exit(EXIT_FAILURE);
    }

  }
  
}
